import os
import numpy as np
import pandas as pd
import umap
from sklearn.preprocessing import StandardScaler
import matplotlib.pyplot as plt
import seaborn as sns
from pathlib import Path

RANDOM_STATE = np.random.RandomState(42)

CATEGORIES = ["Inliers", "Outlier"]
COLORS = ["C0", "C3"]
ZORDER = [1, 2]
SCATTER_STYLE = dict(s=10, edgecolors="w", linewidths=0.2, rasterized=True)
feature_path = "/home/mwshi/project/MassiveQC/Features/features.parquet"

def outlier_umap(feature_path, inliers):
    feature_df = pd.read_parquet(feature_path).dropna()
    scaled_features = scale_feature(feature_df)
    umap_df = umap_feature(scaled_features)
    svg = Path(feature_path).parent / "umap_outliers.svg"
    plot_umap(umap_df, inliers, svg)



def scale_feature(feature_df):
    scaler = StandardScaler()
    scale_feature_df = pd.DataFrame(scaler.fit_transform(feature_df),
                                    index=feature_df.index, columns=feature_df.columns)
    return scale_feature_df


def umap_feature(scaled_features):
    X = scaled_features.values
    reducer = umap.UMAP(random_state=RANDOM_STATE)
    embeddings = reducer.fit_transform(X)
    umap_df = pd.DataFrame(
        embeddings, columns=["UMAP1", "UMAP2"], index=scaled_features.index
    )
    return umap_df


def plot_umap(umap_df, inliers, svg):
    umap_df['labels'] = umap_df.index.to_series().apply(
        lambda x: "Inliers" if x in inliers else "Outlier"
    )
    for cat, color, zorder in zip(CATEGORIES, COLORS, ZORDER):
        df = umap_df.query(f"labels == '{cat}'")
        plt.scatter(df.UMAP1, df.UMAP2, c=color, label=cat, zorder=zorder, **SCATTER_STYLE)

    ax = plt.gca()
    ax.set(xlabel="UMAP 1", ylabel="UMAP 2")
    plt.legend(loc="upper left")
    plt.savefig(svg)


NAME_MAPPER = {
    "percent_alignment": r"Reads Aligned (%)",
    "percent_duplication": f"Duplicated Reads (%)",
    "percent_genes_on": f"Gene Expressed (%)",
    "number_junction_reads": r"Reads at Exon Junctions (#)",
    "percent_utr_bases": r"UTR Bases (%)",
    "gene_body_three_prime": "3' Gene Body Coverage (avg)",
    "percent_intergenic_bases": r"Intergenic Bases (%)",
    "percent_reverse": r"Reverse Oriented Reads (%)",
    "number_genic_reads": "Genic Reads (#)",
    "gene_body_middle": "Middle Gene Body Coverage (avg)",
    "number_reads": "Total Reads (#)",
    "median_cv_coverage": "Coef Var Coverage (median)",
    "gene_body_five_prime": "5' Gene Body Coverage (avg)",
    "percent_mrna_bases": r"mRNA Bases (%)",
    "reads_MQ0": "Reads 0 Mapping Quality (#)",
    "percent_rrna_reads": r"rRNA Reads (%)",
    "number_junctions_on": "Exon Junctions Touched (#)",
    "percent_intronic_bases": r"Intronic Bases (%)",
    "number_multimapping_reads": "Multimapping Reads (#)",
    "average_quality": "Quality Score (avg)",
    "number_reads_too_short": "Reads Too Short (#)",
}


def plot_importance(iso, feature_path):
    mean_shap, columns = iso.mean_shap_values_outliers
    cols = [NAME_MAPPER[x] for x in columns]
    fig, ax = plt.subplots(figsize=plt.figaspect(1.2))
    sns.barplot(x=mean_shap, y=cols, color="C0", ax=ax)
    ax.set(xlabel="mean(|SHAP Value|)", ylabel="")
    svg = Path(feature_path).parent / "feature_importance.svg"
    plt.savefig(svg, bbox_inches='tight')


import logging
from pathlib import Path
import pandas as pd
from .command import run_command
from .parser import remove_file
import numpy as np

logger = logging.getLogger("MassiveQC")


class FeatureCounts(object):
    """Run FeatureCounts and extract count features"""
    def __init__(self, feature_path: str, SRR: str, Bam_dir: str, Count_dir: str,
                 gtf: str, THREADS: int):
        self.feature_path = Path(feature_path)
        self.layout = self.feature_path / "layout" / f"{SRR}.parquet"
        self.strand = self.feature_path / "strand" / f"{SRR}.parquet"
        self.SRR = SRR
        self.Bam_dir = Path(Bam_dir)
        self.Count_dir = Path(Count_dir)
        self.gtf = gtf
        self.THREADS = THREADS

    def FeatureCounts(self):
        counts, jcounts = self.run_featureCounts()
        self.summarize(counts, jcounts)
        return (self.Bam_dir / f"{self.SRR}.sorted.bam").as_posix()

    def run_featureCounts(self):
        # Look up Layout
        layout_ = pd.read_parquet(self.layout).layout[0]
        if layout_ == "PE":
            params = "-p -P -C -J -B "
        else:
            params = "-J "

        # Look up strand
        strand_ = pd.read_parquet(self.strand).strand[0]
        if strand_ == "same_strand":
            params += "-s 1"
        elif strand_ == "opposite_strand":
            params += "-s 2"
        else:
            params += "-s 0"

        counts = self.Count_dir / f"{self.SRR}.counts"
        jcounts = self.Count_dir / f"{self.SRR}.counts.jcounts"
        summary = self.Count_dir / f"{self.SRR}.counts.summary"
        bam = self.Bam_dir / f"{self.SRR}.sorted.bam"

        cmd = (
            f"featureCounts  -T {self.THREADS} {params} "
            f"-a {self.gtf} -o {counts} "
            f"{bam}"
        )
        logger.info(f"{cmd}")
        logger.info(f"{self.SRR} start featureCounts")
        _result = run_command(cmd)
        self._check_log(_result, self.SRR)
        remove_file(summary.as_posix())
        logger.info(f"{self.SRR} complete featureCounts")
        return counts, jcounts

    def summarize(self, counts: Path, jcounts: Path):
        gene_counts = self._get_counts(counts)
        genic_reads = gene_counts.sum()
        percent_genes_on = (gene_counts > 0).mean() * 100

        junction_counts = self._get_counts(jcounts)
        junction_reads = junction_counts.sum()
        number_junctions_on = junction_counts.shape[0]

        df = pd.DataFrame(
            [[genic_reads, percent_genes_on, junction_reads, number_junctions_on]],
            columns=[
                "number_genic_reads",
                "percent_genes_on",
                "number_junction_reads",
                "number_junctions_on",
            ],
            index=pd.Index([self.SRR], name="srr"),
        )
        output_file = self.feature_path / "count_summary" / f"{self.SRR}.parquet"
        df.to_parquet(output_file)

    @staticmethod
    def _get_counts(file_name: Path) -> pd.DataFrame:
        col_name = pd.read_table(file_name, comment="#", nrows=1).columns[-1]
        return pd.read_table(file_name, comment="#", usecols=[col_name], dtype=np.int64).values

    @staticmethod
    def _check_log(log: str, SRR) -> None:
        if not (
                "Read assignment finished" in log or "Summary of counting results" in log
        ):
            raise FeatureCountsException(f"{SRR} not complete")


class FeatureCountsException(Exception):
    """Fastq Screen Processing Exception"""



"""Extract SRA file into Gziped FASTQs"""
import logging
import os
from pathlib import Path
from pysradb.sraweb import SRAweb
from .command import run_command

db = SRAweb()
logger = logging.getLogger("MassiveQC")


class DownloadException(Exception):
    """Basic exception for problems downloading from SRA"""


def _find_aspera_keypath(ascp_key=None):
    """Locate aspera key.
    Parameters
    ----------
    aspera_dir: str
                Location to aspera directory (optional)
    Returns
    -------
    aspera_keypath: str
                    Location to aspera key
    """
    if ascp_key is None:
        aspera_dir = os.path.join(os.path.expanduser("~"), ".aspera")
        aspera_keypath = os.path.join(
            aspera_dir, "connect", "etc", "asperaweb_id_dsa.openssh"
        )
    else:
        aspera_keypath = ascp_key
    if os.path.isfile(aspera_keypath):
        return aspera_keypath


def verify_sra_download(log, SRR):
    if "error" in log.lower():
        logger.warning("Warning: ascp download failed, start download with wget")
        raise DownloadException("Download Failed")
    if "failed" in log.lower() or "unable" in log.lower():
        logger.error("wget download also failed")
        raise DownloadException(f"{SRR} Download Failed")
    if "command not found" in log.lower():
        raise DownloadException("Failed ascp not install")


def sra_ascp(SRR: str, download_path: str, ascp_key) -> Path:
    ascp = "ascp -k1 -T -l 300m -P33001 -i"
    record = db.sra_metadata(SRR, detailed=True)
    record = record.dropna(axis=1, how="any")
    for a in record.to_dict('records'):
        if SRR == a["run_accession"]:
            record = a
    ena_cols = [x for x in list(record.keys()) if "ena_fastq_ftp" in x]
    fastq_col = [x for x in list(record.keys()) if "ena_fastq_http" in x]
    if len(ena_cols) == 0:
        logger.warning(f"{SRR} can not find fastq file on EBI")
        raise DownloadException(f"{SRR} can not find fastq file on EBI")
    filenum = 0
    for ena, fastq in zip(ena_cols, fastq_col):
        try:
            download_url = record[ena]
            if download_url == "":
                continue
            if filenum == 0:
                logger.info(f"{SRR} first file start download")
            else:
                logger.info(f"{SRR} second file start download")
            cmd = "{} {} {} {}".format(
                ascp, _find_aspera_keypath(ascp_key), download_url, download_path
            )
            logger.info(f"running {cmd}")
            log = run_command(cmd)
            verify_sra_download(log, SRR)
        except:
            download_url = record[fastq]
            if download_url == "":
                continue
            if filenum == 0:
                logger.info(f"{SRR} first file start download")
            else:
                logger.info(f"{SRR} second file start download")
            cmd = "wget -N -c -q --timeout=120 -P {} {}".format(
                download_path, download_url
            )
            logger.info(f"running {cmd}")
            log = run_command(cmd)
            verify_sra_download(log, SRR)
        filenum += 1


def get_sra(SRR: str, download_path: str, ascp_key=None) -> Path:
    """Download sra fastq

    **parameter**
    SRR: str
        SRR ID.
    download_path: str
        Download directory
    ascp_key: str
        Location to aspera directory (optional)

    **return**
    Path
    """
    pe_r1 = Path(download_path) / f"{SRR}_1.fastq.gz"
    pe_r2 = Path(download_path) / f"{SRR}_2.fastq.gz"
    se_r1 = Path(download_path) / f"{SRR}.fastq.gz"
    if Path(se_r1).exists():
        logger.info("The file already exists")
        return
    if Path(pe_r1).exists() and Path(pe_r2).exists():
        logger.info("The files already exists")
        return
    try:
        sra_ascp(SRR, download_path, ascp_key)
    except:
        srrs = os.listdir(download_path)
        for srr_file in srrs:
            if SRR in srr_file:
                os.unlink(os.path.join(download_path, srr_file))
        raise DownloadException(f"{SRR} download failed")



from pathlib import Path
import pandas as pd
import logging
from .command import run_command
from .parser import parse_fastq_screen

logger = logging.getLogger("MassiveQC")


def run_fastq_screen(config_file, feature_path, QC_dir, summary_file, SRR, THREADS=1):
    feature_screen = Path(feature_path) / "fastq_screen"
    layout_ = pd.read_parquet(summary_file).layout[0]
    QC_dir = Path(QC_dir)
    if layout_ == "PE" or layout_ == "keep_R1":
        fastq = QC_dir / f"{SRR}_1.fastq.gz"
        feature_file = feature_screen / f"{SRR}_1_screen.txt"
    elif layout_ == "keep_R2":
        fastq = QC_dir / f"{SRR}_2.fastq.gz"
        feature_file = feature_screen / f"{SRR}_1_screen.txt"
    else:
        fastq = QC_dir / f"{SRR}.fastq.gz"
        feature_file = feature_screen / f"{SRR}_screen.txt"

    screen(config_file, feature_screen, fastq, THREADS)
    output_file = feature_screen / f"{SRR}.parquet"
    summarize(feature_file, output_file, SRR)
    feature_file.unlink()
    (feature_screen / f"{feature_file.stem}.html").unlink()
    (feature_screen / f"{feature_file.stem}.png").unlink()


def screen(config_file, feature_screen, fastq, THREADS: int) -> None:
    cmd = f"fastq_screen --outdir {feature_screen} " \
          f"--force --aligner bowtie2 --threads {THREADS} " \
          f"--conf {config_file} " \
          f"--subset 100000 " \
          f"{fastq} "
    logger.info(f"running {cmd}")
    log_info = run_command(cmd)
    if "Processing complete" not in log_info:
        logger.error(log_info)
        raise FastqScreenException


def summarize(feature_file: Path, output_file: Path, SRR: str) -> None:
    """Summarizes fastq screen results
        Calculates the number of reads mapping to each specific reference. Ignores reads the map to multiple references.
        Goes from this:
        | reference   |   multiple_hits_multiple_libraries_count |   multiple_hits_multiple_libraries_percent |   multiple_hits_one_library_count |   multiple_hits_one_library_percent |   one_hit_multiple_libraries_count |   one_hit_multiple_libraries_percent |   one_hit_one_library_count |   one_hit_one_library_percent |   reads_processed_count |   unmapped_count |   unmapped_percent |
        |:------------|-----------------------------------------:|-------------------------------------------:|----------------------------------:|------------------------------------:|-----------------------------------:|-------------------------------------:|----------------------------:|------------------------------:|------------------------:|-----------------:|-------------------:|
        | adapters    |                                       48 |                                       0.05 |                                 0 |                                0    |                                  0 |                                 0    |                           0 |                          0    |                   99973 |            99925 |              99.95 |
        | dm6         |                                     1713 |                                       1.71 |                              6278 |                                6.28 |                                224 |                                 0.22 |                       88393 |                         88.42 |                   99973 |             3365 |               3.37 |
        | ecoli       |                                        1 |                                       0    |                                 0 |                                0    |                                  0 |                                 0    |                           2 |                          0    |                   99973 |            99970 |             100    |
        ...
        To this:
        |            |   adapters_pct_reads_mapped |   dm6_pct_reads_mapped |   ecoli_pct_reads_mapped |   ercc_pct_reads_mapped |   hg19_pct_reads_mapped |   phix_pct_reads_mapped |   rRNA_pct_reads_mapped |   wolbachia_pct_reads_mapped |   yeast_pct_reads_mapped |
        |:-----------|----------------------------:|-----------------------:|-------------------------:|------------------------:|------------------------:|------------------------:|------------------------:|-----------------------------:|-------------------------:|
        | SRR0000001 |                           0 |                94.6966 |               0.00200054 |                       0 |               0.0160043 |                       0 |              0.00100027 |                            0 |               0.00500135 |
    """
    logger.info(F"Start extract {SRR} fastq screen result")
    df = parse_fastq_screen(feature_file).set_index("reference").fillna(0)
    summarized = (
        (
                (df.one_hit_one_library_count + df.multiple_hits_one_library_count)
                / df.reads_processed_count
                * 100
        )
            .rename(SRR)
            .rename_axis("")
            .to_frame()
            .T.rename_axis("srr")
    )
    summarized.columns = [f"{col}_pct_reads_mapped" for col in summarized.columns]
    summarized.to_parquet(output_file)
    logger.info("Complete, remove fastq screen result files")



class FastqScreenException(Exception):
    """Fastq Screen Processing Exception"""


def fastq_screen(SRR, QC_dir, feature_path, config_file, THREADS):
    """Use fastqscreen to identify RNA seq quality"""
    logger.info(f"Start fastq screen {SRR}")
    layout_file = Path(feature_path) / "layout" / f"{SRR}.parquet"
    try:
        run_fastq_screen(config_file, feature_path, QC_dir, layout_file, SRR, THREADS)
        logger.info(f"Complete fast_screen {SRR} fastq file")
    except FastqScreenException:
        logger.warning(f"{SRR}: fastq screen did not complete")
        raise FastqScreenException



import logging
from pathlib import Path
import pandas as pd
from .command import run_command
from typing import Optional
from .parser import parse_picard_markduplicate_metrics, remove_file
import numpy as np

logger = logging.getLogger("MassiveQC")


class MarkDuplicates(object):
    """run picard MarkDuplicates and extract the result"""
    def __init__(self, feature_path: str, SRR: str, Bam_dir: str,
                 THREADS: int, picard: str, MEM: Optional[int] = 3):
        self.feature_path = Path(feature_path)
        self.layout = Path(feature_path) / f"{SRR}.parquet"
        self.SRR = SRR
        self.Bam_dir = Path(Bam_dir)
        self.THREADS = THREADS
        self.MEM = MEM
        self.picard = picard

    def markduplicates(self):
        bam = self.Bam_dir / f"{self.SRR}.sorted.bam"
        metrics = self.run_markduplicates(bam)
        self.summarize(metrics)

    def run_markduplicates(self, bam: Path):
        dedup_bam = self.Bam_dir / f"{self.SRR}.dedup.bam"
        metrics = self.feature_path / f"{self.SRR}.metrics"
        cmd = (
            f"java -Xmx{self.MEM}g "
            f"-jar {self.picard} MarkDuplicates "
            f"INPUT={bam} "
            f"OUTPUT={dedup_bam} "
            f"METRICS_FILE={metrics}"
        )
        logger.info(f"{cmd}")
        logger.info(f"{self.SRR} Start markduplicates")
        _result = run_command(cmd)
        self._check_log(_result, self.SRR)
        remove_file(dedup_bam.as_posix())
        return metrics

    def summarize(self, metrics: Path):
        dtypes = {
            "UNPAIRED_READS_EXAMINED": np.int64,
            "READ_PAIRS_EXAMINED": np.int64,
            "UNPAIRED_READ_DUPLICATES": np.int64,
            "READ_PAIR_DUPLICATES": np.int64,
            "PERCENT_DUPLICATION": np.float64,
            "ESTIMATED_LIBRARY_SIZE": np.int64,
        }
        df = parse_picard_markduplicate_metrics(metrics)[dtypes.keys()].fillna(0).astype(dtypes)
        df.PERCENT_DUPLICATION = df.PERCENT_DUPLICATION * 100
        df.columns = [col.lower() for col in df.columns]
        df.index = pd.Index([self.SRR], name="srr")
        output_file = self.feature_path / "markduplicates" / f"{self.SRR}.parquet"
        remove_file(metrics.as_posix())
        df.to_parquet(output_file)

    @staticmethod
    def _check_log(log: str, SRR) -> None:
        if "MarkDuplicates done" not in log:
            raise PicardException(f"{SRR} not complete")


class PicardException(Exception):
    """Picard Processing Exception"""



from pathlib import Path
from multiprocessing import Pool
import os, argparse
import pandas as pd

THREADS = 3

# NOTE: features commented out are being dropped because they are repetitive or not important.
FEATURE_AGG = {
    "rRNA_pct_reads_mapped": "mean",
    "too_short": "sum",
    "num_reads": "sum",
    "num_multimappers": "sum",
    "per_alignment": "mean",
    "reads_MQ0": "sum",
    "average_quality": "mean",
    "Percent Reverse": "mean",
    "percent_utr_bases": "mean",  # Exclusive of percent_coding and correlated with percent_mrna
    "percent_intronic_bases": "mean",
    "percent_intergenic_bases": "mean",
    "percent_mrna_bases": "mean",
    "median_cv_coverage": "mean",
    "percent_duplication": "mean",
    "number_genic_reads": "sum",
    "percent_genes_on": "mean",
    "number_junction_reads": "sum",
    "number_junctions_on": "sum",
    "gene_body_five_prime": "mean",
    "gene_body_middle": "mean",
    "gene_body_three_prime": "mean",
}

FEATURE_RENAME = {
    "rRNA_pct_reads_mapped": "percent_rrna_reads",
    "too_short": "number_reads_too_short",
    "num_reads": "number_reads",
    "num_multimappers": "number_multimapping_reads",
    "per_alignment": "percent_alignment",
    "Percent Reverse": "percent_reverse",
}


def build_features(input: str, Features: Path):
    """Create a feature set for machine learning.
    Identify and munge a set of features from the files generated by the prealn-wf
    and aln-wf.
    Features include:
    * CollectRNASeqMetrics
        * PCT_CODING_BASES
        * PCT_UTR_BASES
        * PCT_INTRONIC_BASES
        * PCT_INTERGENIC_BASES
        * PCT_MRNA_BASES
        * MEDIAN_CV_COVERAGE
        * MEDIAN_5PRIME_BIAS
        * MEDIAN_3PRIME_BIAS
    * CollectRNASeqMetrics Gene Body Coverage
    * Markduplicates
        * PERCENT_DUPLICATION
    * Fastq Screen
        * Percent reads mapping to rRNA.
    * FeatureCounts
        * Number of reads mapping to junction

    **parameter**
    input: str
        The input file with srx and srr
    Features: Path
        The Features directory

    **return**
    None
    """
    feature_dict = {
        "layout": Features / "layout.parquet",
        "fastq_screen": Features / "fastq_screen.parquet",
        "atropos": Features / "atropos.parquet",
        "hisat2": Features / "hisat2.parquet",
        "aln_stats": Features / "aln_stats.parquet",
        "rnaseqmetrics": Features / "rnaseqmetrics.parquet",
        "genebody_coverage": Features / "genebody_coverage.parquet",
        "markduplicates": Features / "markduplicates.parquet",
        "count_summary": Features / "count_summary.parquet"
    }
    done_sample_file = Features / "done_sample.txt"
    srr_df = pd.read_table(input, comment="#")
    if len(srr_df.columns) == 1:
        # only have srr column
        srr_df.columns = ["srr"]
    elif len(srr_df.columns) == 2:
        # have srx and srr column
        srr_df.columns = ["srx", "srr"]
    srr_df = srr_df.set_index("srr", drop=False)
    done_sample_df = pd.read_table(done_sample_file, comment="#")
    done_srrs = done_sample_df["srr"].values.tolist()
    if len(srr_df.columns) == 2:
        (
            workflow_data(done_srrs, feature_dict)
                .join(srr_df)
                .pipe(aggregate_gene_body_coverage)
                .groupby("srx")
                .agg(FEATURE_AGG)
                .rename(columns=FEATURE_RENAME)
                .to_parquet(Features / "features.parquet")
        )
    else:
        (
            workflow_data(done_srrs, feature_dict)
                .join(srr_df)
                .pipe(aggregate_gene_body_coverage)
                .loc[:, FEATURE_AGG.keys()]
                .rename(columns=FEATURE_RENAME)
                .to_parquet(Features / "features.parquet")
        )


def workflow_data(srrs: list, workflow_folders: dict) -> pd.DataFrame:
    pool = Pool(THREADS)
    df = (
        pd.concat(
            pool.map(pd.read_parquet, [path for _, path in workflow_folders.items()]),
            axis=1,
            sort=False,
        )
            .rename_axis("srr")
            .reindex(srrs)
    )
    pool.close()

    return df


def aggregate_gene_body_coverage(df: pd.DataFrame) -> pd.DataFrame:
    """Sum gene body coverage to tertile.
    GBC is reported as a centile, with positions next to each other being
    highly correlated. For machine learning, I am aggregating these features
    into a tertiles (i.e., 5', middle, 3').
    """
    cols = [f"pos_{i}" for i in range(101)]
    five_prime, middle, three_prime = cols[:33], cols[33:68], cols[68:]
    return (
        df.assign(gene_body_five_prime=lambda x: x[five_prime].sum(axis=1))
            .assign(gene_body_middle=lambda x: x[middle].sum(axis=1))
            .assign(gene_body_three_prime=lambda x: x[three_prime].sum(axis=1))
            .drop(cols, axis=1)
    )


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', required=True, type=str,
                        help='Input file, containing two columns srx and srr')
    parser.add_argument('-o', '--outdir', required=True, type=str,
                        help="Path to result output directory of main process.")
    args = parser.parse_args()
    Features = Path(args.outdir) / "Features"
    build_features(args.input, Features)



from pathlib import Path
import pandas as pd
import logging
from typing import Tuple
import re
from .command import run_command

logger = logging.getLogger("MassiveQC")


def atropos(feature_path: str, SRR: str, QC_dir: str, THREADS: int):
    """This function is main part which is used to trim the reads.
    Filter reads that are less than 25bp.

    **parameter**
    feature_path: str
        The directory of Features.
    SRR: str
        SRR ID.
    QC_dir: str
       The directory of result fastq file after check_fq.
    THREADS: int
        Thread number for atropos.
    **return**
    None
    """
    logger.info(f"Start atropos {SRR}")
    try:
        QC_dir = Path(QC_dir)
        layout = Path(feature_path) / "layout" / f"{SRR}.parquet"
        layout_ = pd.read_parquet(layout).layout[0]
        results = run_atropos(layout_, SRR, QC_dir, THREADS)
        output = Path(feature_path) / "atropos" / f"{SRR}.parquet"
        summarize(results, output, SRR)
        if layout_ == "PE":
            r1_trim = QC_dir / f"{SRR}_1.trim.fastq.gz"
            r2_trim = QC_dir / f"{SRR}_2.trim.fastq.gz"
            if r1_trim.exists() and r2_trim.exists():
                (QC_dir / f"{SRR}_1.fastq.gz").unlink()
                (QC_dir / f"{SRR}_2.fastq.gz").unlink()
        elif layout_ == "Keep_R1":
            r1_trim = QC_dir / f"{SRR}_1.trim.fastq.gz"
            if r1_trim.exists():
                (QC_dir / f"{SRR}_1.fastq.gz").unlink()
        elif layout_ == "Keep_R2":
            r1_trim = QC_dir / f"{SRR}_2.trim.fastq.gz"
            if r1_trim.exists():
                (QC_dir / f"{SRR}_2.fastq.gz").unlink()
        else:
            r1_trim = QC_dir / f"{SRR}.trim.fastq.gz"
            if r1_trim.exists():
                (QC_dir / f"{SRR}.fastq.gz").unlink()
        logger.info(f"Complete atropos {SRR}")
    except AtroposException as error:
        logger.warning(f"Flagging {SRR} as Atropos Bad")
        raise AtroposException("Atropos Bad")


def run_atropos(layout_, SRR, QC_dir: Path, THREADS) -> str:
    """Run the atropos command in shell"""
    # adapters = os.path.join(os.path.abspath(__file__), "sequencing_adapters.fa")
    if layout_ == "PE":
        r1 = QC_dir / f"{SRR}_1.fastq.gz"
        r2 = QC_dir / f"{SRR}_2.fastq.gz"
        r1_trim = QC_dir / f"{SRR}_1.trim.fastq.gz"
        r2_trim = QC_dir / f"{SRR}_2.trim.fastq.gz"
        cmd = f"atropos trim " \
              f"-q 20 --minimum-length 25 " \
              f"--threads {THREADS} " \
              f"-pe1 {r1} -pe2 {r2} -o {r1_trim} -p {r2_trim}"
    elif layout_ == "Keep_R1":
        r1 = QC_dir / f"{SRR}_1.fastq.gz"
        r1_trim = QC_dir / f"{SRR}_1.trim.fastq.gz"
        cmd = f"atropos trim " \
              "-q 20 -U 0 --minimum-length 25 " \
              f"--threads {THREADS} " \
              f"-se {r1} -o {r1_trim}"
    elif layout_ == "Keep_R2":
        r2 = QC_dir / f"{SRR}_2.fastq.gz"
        r2_trim = QC_dir / f"{SRR}_2.trim.fastq.gz"
        cmd = f"atropos trim " \
              "-q 20 -U 0 --minimum-length 25 " \
              f"--threads {THREADS} " \
              f"-se {r2} -o {r2_trim}"
    else:
        r1 = QC_dir / f"{SRR}.fastq.gz"
        r1_trim = QC_dir / f"{SRR}.trim.fastq.gz"
        cmd = f"atropos trim " \
              "-q 20 -U 0 --minimum-length 25 " \
              f"--threads {THREADS} " \
              f"-se {r1} -o {r1_trim}"
    logger.info(f"running {cmd}")
    results = run_command(cmd, verbose=True)
    return results


def summarize(log: str, output: Path, SRR: str) -> None:
    """Extract features from atropos result"""
    df = pd.DataFrame(
        [parse_atropos_log(log, SRR)],
        index=pd.Index([SRR], name="srr"),
        columns=["total_processed", "total_written", "too_short"],
    )
    if df.total_written[0] < 1_000:
        raise AtroposException("<1,000 reads")
    df.to_parquet(output)


def parse_atropos_log(log_text: str, SRR: str) -> Tuple[int, int, int]:
    """Parse atropos log information
       Example Atropos output
        # Example of SE
        Reads                                 records  fraction
        ----------------------------------- --------- ---------
        Total reads processed:                 85,319
        Reads with adapter:                     4,725      5.5%
        Reads that were too short:              1,879      2.2%
        Reads written (passing filters):       83,440     97.8%
        # Example of PE
        Pairs                                records fraction
        ----------------------------------- -------- --------
        Total read pairs processed:            1,354
        Read 1 with adapter:                   106     7.8%
        Read 2 with adapter:                   123     9.1%
        Pairs that were too short:                16     1.2%
        Pairs written (passing filters):       1,338    98.8%
    """
    if "ERROR" in log_text:
        logger.error(log_text)
        logger.warning(f"{SRR}: Atropos reported an error")
    try:
        log_text = log_text.replace(",", "")
        tot_processed = int(re.findall(r"Total read.*processed:\s+(\d+)", log_text)[0])
        tot_written = int(
            re.findall(r"[Read|Pair]s written \(passing filters\):\s+(\d+)", log_text)[0]
        )
        too_short = int(re.findall(r"[Read|Pair]s that were too short:\s+(\d+)", log_text)[0])
        return tot_processed, tot_written, too_short
    except IndexError:
        raise AtroposException("Unable to parse log")


class AtroposException(Exception):
    """Basic Atropos Exception"""



import re
from collections import OrderedDict
import numpy as np
import pandas as pd
from pathlib import Path
from io import StringIO
def parse_fastq_screen(fname):
    """Parser for fastq screen.
    Adapted from multiqc.
    Returns
    -------
    pandas.DataFrame: A single row dataframe.
    """
    with open(fname, "r") as fh:
        header = ["reference", "type", "value"]
        parsed = []

        for l in fh:
            fqs = re.search(
                r"^(\S+)\s+(\d+)\s+(\d+)\s+-?([\d\.]+)\s+(\d+)\s+([\d\.]+)\s+(\d+)\s+([\d\.]+)\s+(\d+)\s+([\d\.]+)\s+(\d+)\s+([\d\.]+)$",
                l,
            )
            if fqs:
                org = fqs.group(1)
                parsed.append((org, "reads_processed_count", int(fqs.group(2))))
                parsed.append((org, "unmapped_count", int(fqs.group(3))))
                parsed.append((org, "unmapped_percent", float(fqs.group(4))))
                parsed.append((org, "one_hit_one_library_count", int(fqs.group(5))))
                parsed.append((org, "one_hit_one_library_percent", float(fqs.group(6))))
                parsed.append((org, "multiple_hits_one_library_count", int(fqs.group(7))))
                parsed.append((org, "multiple_hits_one_library_percent", float(fqs.group(8))))
                parsed.append((org, "one_hit_multiple_libraries_count", int(fqs.group(9))))
                parsed.append((org, "one_hit_multiple_libraries_percent", float(fqs.group(10))))
                parsed.append((org, "multiple_hits_multiple_libraries_count", int(fqs.group(11))))
                parsed.append(
                    (org, "multiple_hits_multiple_libraries_percent", float(fqs.group(12)))
                )

        if len(parsed) == 0:
            return None
        else:
            df = pd.DataFrame(parsed, columns=header)
            udf = df.set_index(["reference", "type"]).unstack()
            udf.columns = udf.columns.droplevel()
            return udf.reset_index()


def parse_hisat2(log):
    """Parse hisat2."""
    parsed = OrderedDict(
        [
            ("num_reads", np.nan),
            ("num_reads_paired", np.nan),
            ("num_reads_unpaired", np.nan),
            ("num_concordant_reads_unaligned", np.nan),
            ("num_concordant_reads_uniquely_aligned", np.nan),
            ("num_concordant_multimappers", np.nan),
            ("num_discordant_reads_aligned", np.nan),
            ("num_unaligned", np.nan),
            ("num_uniquely_aligned", np.nan),
            ("num_multimappers", np.nan),
            ("per_alignment", np.nan),
        ]
    )

    header = {
        "reads; of these:": "num_reads",
        "were paired; of these:": "num_reads_paired",
        "were unpaired; of these:": "num_reads_unpaired",
        "aligned concordantly 0 times": "num_concordant_reads_unaligned",
        "aligned concordantly exactly 1 time": "num_concordant_reads_uniquely_aligned",
        "aligned concordantly >1 times": "num_concordant_multimappers",
        "aligned discordantly 1 time": "num_discordant_reads_aligned",
        "aligned 0 times": "num_unaligned",
        "aligned exactly 1 time": "num_uniquely_aligned",
        "aligned >1 times": "num_multimappers",
        "overall alignment rate": "per_alignment",
    }

    for l in log.split("\n"):
        row = l.strip()
        if row.startswith("Warning") or row.startswith("---"):
            continue
        fqs = re.search(r"^([\d\.]+)[%]*\s([\(\)\d\.%]*\s)?(.*)$", row)
        if fqs:
            name = fqs.group(3)
            value = fqs.group(1)
            if name in header:
                head = header[name]
                parsed[head] = float(value)

    if len(parsed) == 0:
        return None
    else:
        return pd.DataFrame(parsed, index=[0])

def parse_samtools_stats(log):
    """Parse rseqc samtools stats."""
    with StringIO(log) as fh:
        parsed = OrderedDict()
        for l in fh:
            if l.startswith("SN"):
                fqs = re.search(r"^SN\s+(.+?):\s+([\d\.]+)\s.*$", l)
                if fqs:
                    name = fqs.group(1).replace(" ", "_")
                    value = fqs.group(2)
                    if ("." in value) or ("average" in name):
                        parsed[name] = float(value)
                    else:
                        parsed[name] = int(value)
        if len(parsed) == 0:
            return None
        else:
            return pd.DataFrame(parsed, index=[0])

def parse_bamtools_stats(log):
    """Parse bamtools stats."""
    with StringIO(log) as fh:
        parsed = OrderedDict()
        for l in fh:
            fqs = re.search(r"^(.+?):\s+(\d+).*$", l)
            if fqs:
                parsed[fqs.group(1).replace("'", "")] = int(fqs.group(2))
        if len(parsed) == 0:
            return None
        else:
            df = pd.DataFrame(parsed, index=[0])
            df["Percent Mapped"] = df["Mapped reads"] / df["Total reads"] * 100
            df["Percent Forward"] = df["Forward strand"] / df["Total reads"] * 100
            df["Percent Reverse"] = df["Reverse strand"] / df["Total reads"] * 100
            df["Percent Failed QC"] = df["Failed QC"] / df["Total reads"] * 100
            df["Percent Duplicates"] = df["Duplicates"] / df["Total reads"] * 100
            df["Percent Paired-end"] = df["Paired-end reads"] / df["Total reads"] * 100
            return df

def parse_picardCollect_summary(fname):
    """Parser for picard collectRNAMetrics summary."""
    dtypes = {
        "CODING_BASES": np.int64,
        "CORRECT_STRAND_READS": np.int64,
        "IGNORED_READS": np.int64,
        "INCORRECT_STRAND_READS": np.int64,
        "INTERGENIC_BASES": np.int64,
        "INTRONIC_BASES": np.int64,
        "LIBRARY": np.float64,
        "MEDIAN_3PRIME_BIAS": np.float64,
        "MEDIAN_5PRIME_BIAS": np.float64,
        "MEDIAN_5PRIME_TO_3PRIME_BIAS": np.float64,
        "MEDIAN_CV_COVERAGE": np.float64,
        "NUM_R1_TRANSCRIPT_STRAND_READS": np.float64,
        "NUM_R2_TRANSCRIPT_STRAND_READS": np.float64,
        "NUM_UNEXPLAINED_READS": np.float64,
        "PCT_CODING_BASES": np.float64,
        "PCT_CORRECT_STRAND_READS": np.float64,
        "PCT_INTERGENIC_BASES": np.float64,
        "PCT_INTRONIC_BASES": np.float64,
        "PCT_MRNA_BASES": np.float64,
        "PCT_R1_TRANSCRIPT_STRAND_READS": np.float64,
        "PCT_R2_TRANSCRIPT_STRAND_READS": np.float64,
        "PCT_RIBOSOMAL_BASES": np.float64,
        "PCT_USABLE_BASES": np.float64,
        "PCT_UTR_BASES": np.float64,
        "PF_ALIGNED_BASES": np.int64,
        "PF_BASES": np.int64,
        "READ_GROUP": np.float64,
        "RIBOSOMAL_BASES": np.float64,
        "SAMPLE": np.float64,
        "UTR_BASES": np.int64,
    }
    with open(fname, "r") as fh:
        for l in fh:
            if l.startswith("#"):
                continue
            if l.startswith("PF_BASES"):
                parsed = l
                parsed += next(fh)
                break

        if len(parsed) == 0:
            return None
        else:
            return pd.read_csv(StringIO(parsed), sep="\t", na_values="?", dtype=dtypes)

def parse_picardCollect_hist(fname):
    """Parser for picard collectRNAMetrics summary."""
    parsed = ""
    with open(fname, "r") as fh:
        for l in fh:
            if l.startswith("#"):
                continue
            if l.startswith("normalized"):
                parsed = l
                while True:
                    try:
                        parsed += next(fh)
                    except StopIteration:
                        break
                break

        if len(parsed) == 0:
            return None
        else:
            df = pd.read_csv(StringIO(parsed), sep="\t", index_col=0).T
            df.reset_index(drop=True, inplace=True)
            df.columns = [f"pos_{x}" for x in range(101)]
            return df

def parse_picard_markduplicate_metrics(fname):
    """Parser for picard markduplicates."""
    with open(fname, "r") as fh:
        for l in fh:
            if l.startswith("## METRICS CLASS"):
                dat = next(fh)
                dat += next(fh)
                break
    return pd.read_csv(StringIO(dat), sep="\t", comment="#", na_values="?")

def remove_file(file_name: str):
    if file_name is None:
        return
    pth = Path(file_name)
    if pth.exists() & pth.is_file():
        pth.unlink()


from typing import Tuple
import numpy as np
import pandas as pd
import shap
from sklearn.ensemble import IsolationForest
from sklearn.model_selection import train_test_split


class SraIsolationForest:
    def __init__(
        self,
        features: pd.DataFrame,
        random_state=None,
        iso_kwargs=None,
        explain_kwargs=None,
        shap_values_kwargs=None,
    ):
        """Instantiate, fit, and explain an isolation forest.
        Parameters
        ----------
        features : pd.DataFrame
            Set of features to fit. Automatically does X_train/X_test splitting.
        random_state: np.random.RandomState, optional
            Numpy random state.
        iso_kwargs : optional
            Passed directly to `sklearn.ensemble.IsolationForest, by default None
        explain_kwargs : optional
            Passed directly to `shap.TreeExplainer`, by default None
        shap_values_kwargs : optional
            Passed directly to `shap.TreeExplainer.shap_values`, by default None
        Attributes
        ----------
        features : pd.DataFrame
            Stores the features passed at init.
        X_train : pd.DataFrame
            The subset used to X_train the isolation forest model.
        X_test : pd.DataFrame
            The subset used to X_test
        model_ : sklearn.ensemble.IsolationForest
            Fitted isolation forest model.
        explainer_ : shap.TreeExplainer
            Shap model to explain isolation forest results
        expected_value_ : float
            The mean value from `shap.TreeExpliner.expected_value`.
        index : pd.Index
            features.index
        columns : pd.Columns
            features.columns
        Note that many method names are available as suffixed version
        {_test, _train, _all} to return the results for
        {X_test, X_train, or features}
        Methods
        -------
        predict
            Return an array from running isolation forest outlier detection.
        inliers
            Return a dataframe with inliers.
        outliers
            Retrun a dataframe with outliers.
        isinlier
            Return a boolean array where inlier.
        isoutlier
            Return a boolean array where outlier.
        prop_outliers
            Return the proportion of samples that are outliers.
        shap_values
            Return an array with shap_values
        """
        # Prep features
        self.random_state = random_state or np.random.RandomState(42)
        self.features = features
        self.X_train, self.X_test = train_test_split(self.features, random_state=self.random_state)

        # Run Isolation Forest
        iso_kwargs = iso_kwargs or {}
        self.model_ = IsolationForest(**iso_kwargs).fit(self.X_train)

        # Run Tree Explainer
        explain_kwargs = explain_kwargs or {}
        self.explainer_ = shap.TreeExplainer(self.model_, data=self.inliers_test, **explain_kwargs)
        self.expected_value_ = self.explainer_.expected_value

        shap_values_kwargs = shap_values_kwargs or {}
        self.shap_values_ = self.explainer_.shap_values(self.X_test, **shap_values_kwargs)

    @property
    def index(self) -> pd.Index:
        """Return sample index"""
        return self.features.index

    @property
    def columns(self) -> pd.Index:
        """Return feature names."""
        return self.features.columns

    def predict(self, X: pd.DataFrame) -> np.ndarray:
        """Run Isolation Forest outlier detection.
        Parameters
        ----------
        X : pd.DataFrame
            Data frame of features to run prediction
        Returns
        -------
        np.ndarray
            An array where -1 are outliers and 1 are inliers.
        """
        return self.model_.predict(X)

    def isinlier(self, X: pd.DataFrame) -> np.ndarray:
        """Create inlier boolean mask.
        Parameters
        ----------
        X : pd.DataFrame
            Data frame of features to run prediction
        Returns
        -------
        np.ndarray
            An array where True are inliers and False are outliers.
        """
        return np.where(self.model_.predict(X) == 1, True, False)

    @property
    def isinlier_test(self) -> np.ndarray:
        """Inlier boolean mask for X_test data"""
        return self.isinlier(self.X_test)

    @property
    def isinlier_train(self) -> np.ndarray:
        """Inlier boolean mask for X_train data"""
        return self.isinlier(self.X_train)

    @property
    def isinlier_all(self) -> np.ndarray:
        """Inlier boolean mask for all data"""
        return self.isinlier(self.features)

    def isoutlier(self, X: pd.DataFrame) -> np.ndarray:
        """Create outlier boolean mask.
        Parameters
        ----------
        X : pd.DataFrame
            Data frame of features to run prediction
        Returns
        -------
        np.ndarray
            An array where True are outliers and False are inliers.
        """
        return np.where(self.model_.predict(X) == -1, True, False)

    @property
    def isoutlier_test(self) -> np.ndarray:
        """Inlier boolean mask for X_test data"""
        return self.isoutlier(self.X_test)

    @property
    def isoutlier_train(self) -> np.ndarray:
        """Inlier boolean mask for X_train data"""
        return self.isoutlier(self.X_train)

    @property
    def isoutlier_all(self) -> np.ndarray:
        """Inlier boolean mask for all data"""
        return self.isoutlier(self.features)

    def inliers(self, X: pd.DataFrame) -> pd.DataFrame:
        """Return a data frame of inliers"""
        return X[self.isinlier(X)]

    @property
    def inliers_test(self) -> pd.DataFrame:
        """Return a data frame of inliers"""
        return self.inliers(self.X_test)

    @property
    def inliers_train(self) -> pd.DataFrame:
        """Return a data frame of inliers"""
        return self.inliers(self.X_train)

    @property
    def inliers_all(self) -> pd.DataFrame:
        """Return a data frame of inliers"""
        return self.inliers(self.features)

    def outliers(self, X: pd.DataFrame) -> pd.DataFrame:
        """Return a data frame of outliers"""
        return X[self.isoutlier(X)]

    @property
    def outliers_test(self) -> pd.DataFrame:
        """Return a data frame of inliers"""
        return self.outliers(self.X_test)

    @property
    def outliers_train(self) -> pd.DataFrame:
        """Return a data frame of inliers"""
        return self.outliers(self.X_train)

    @property
    def outliers_all(self) -> pd.DataFrame:
        """Return a data frame of inliers"""
        return self.outliers(self.features)

    def prop_outliers(self, X) -> float:
        """Return the proportion of outliers in X."""
        return np.mean(self.isoutlier(X))

    @property
    def prop_outliers_test(self):
        """Return the proportion of outliers in X_test data."""
        return self.prop_outliers(self.X_test)

    @property
    def prop_outliers_train(self):
        """Return the proportion of outliers in X_train data."""
        return self.prop_outliers(self.X_train)

    @property
    def prop_outliers_all(self):
        """Return the proportion of outliers in all data."""
        return self.prop_outliers(self.features)

    @property
    def shap_values(self):
        """Retruns calculated shap values for X_test data."""
        return self.shap_values_

    @property
    def shap_values_inliers(self):
        """Return shap values for inliers for X_test data"""
        return self.shap_values_[self.isinlier_test]

    @property
    def shap_values_outliers(self):
        """Return shap values for outliers for X_test data"""
        return self.shap_values_[self.isoutlier_test]

    def mean_shap_values(self, shap_values) -> Tuple[np.ndarray, pd.Index]:
        """Calculate the mean(abs(shap_values)).
        Returns
        -------
        Tuple[np.ndarray, pd.Index]
            Mean absolute shapely values and feature names ordered by
            importance.
        """
        mean_shap = np.mean(np.abs(shap_values), axis=0)
        idx_order = np.argsort(mean_shap)[::-1]
        return mean_shap[idx_order], self.columns[idx_order]

    @property
    def mean_shap_values_inliers(self) -> Tuple[np.ndarray, pd.Index]:
        """Calculate the mean(abs(shap_values)).
        Returns
        -------
        Tuple[np.ndarray, pd.Index]
            Mean absolute shapely values and feature names ordered by
            importance.
        """
        return self.mean_shap_values(self.shap_values[self.isinlier_test])

    @property
    def mean_shap_values_outliers(self) -> Tuple[np.ndarray, pd.Index]:
        """Calculate the mean(abs(shap_values)).
        Returns
        -------
        Tuple[np.ndarray, pd.Index]
            Mean absolute shapely values and feature names ordered by
            importance.
        """
        return self.mean_shap_values(self.shap_values[self.isoutlier_test])


import argparse
from pathlib import Path

from .feature_store import check_done_sample, feature_store
from .build_features import build_features
from .detection import detection



def main():
    """Combine multiple modules to provide commands for cli"""
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', required=True, type=str, help='Input file, containing two columns srx and srr')
    parser.add_argument('-o', '--outdir', required=True, type=str, help="Path to result output directory of main process.")
    args = parser.parse_args()
    done_samples = check_done_sample(args.outdir)
    feature_store(done_samples, args.outdir)
    done_samples = check_done_sample(args.outdir)
    feature_store(done_samples, args.outdir)
    Features = Path(args.outdir) / "Features"
    build_features(args.input, Features)
    features_file = (Features / "features.parquet").as_posix()
    detection(features_file)


if __name__ == "__main__":
    main()


"""Aggregate pre-alignment workflow data"""
import argparse
import os
from pathlib import Path
from typing import Optional, Set, Tuple

import pandas as pd


def feature_store(done_samples: list, outdir: str):
    """Merge the features of each sample"""
    OUTPUTS = [
        "aln_stats",
        "atropos",
        "count_summary",
        "fastq_screen",
        "genebody_coverage",
        "hisat2",
        "markduplicates",
        "rnaseqmetrics",
        "strand",
        "layout"
    ]
    PREALN_OUTPUT = Path(outdir) / "Features"
    for output in OUTPUTS:
        print(f"Aggregating: {output:>20}", end="\t")
        aggregate_data_store(
            set(done_samples), PREALN_OUTPUT / output, PREALN_OUTPUT / f"{output}.parquet"
        )


def aggregate_data_store(workflow_samples: set, data_folder_pth: Path, data_store_pth: Path):
    data_store, old_samples = load_data_store(data_store_pth)
    new_samples = find_new_samples(workflow_samples, old_samples, data_folder_pth)

    print(f"({len(new_samples):,})")
    new_data = load_data_folder(new_samples, data_folder_pth)

    if new_data is not None:
        updated_data = update_data_store(data_store, new_data)
        updated_data.to_parquet(data_store_pth)


def load_data_store(data_store_pth: Path) -> Tuple[Optional[pd.DataFrame], Set[str]]:
    if data_store_pth.exists():
        data_store = pd.read_parquet(data_store_pth)
        old_samples = set(data_store.index.unique())
    else:
        data_store, old_samples = None, set()
    return data_store, old_samples


def find_new_samples(workflow_samples: set, old_samples: set, data_folder_pth: Path) -> Set[str]:
    dir_content = {file_name.stem for file_name in data_folder_pth.iterdir()}
    return workflow_samples.intersection(dir_content - old_samples)


def load_data_folder(samples: set, data_pth: Path) -> Optional[pd.DataFrame]:
    if not samples:
        return None

    return pd.concat(
        [pd.read_parquet(data_pth / f"{sample}.parquet") for sample in samples], sort=False
    )


def update_data_store(data_store: Optional[pd.DataFrame], new_data: pd.DataFrame) -> pd.DataFrame:
    if data_store is not None:
        return pd.concat([data_store, new_data], sort=False)
    return new_data


def check_done_sample(outdir):
    done_sample_file = Path(outdir) / "Features" / "done_sample.txt"
    done_sample_dir = Path(outdir) / "Features" / "DoneSample"
    done_samples = os.listdir(done_sample_dir)
    pd.DataFrame(done_samples, columns=["srr"]).to_csv(done_sample_file, sep ='\t')
    return done_samples


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-i', '--input', required=True, type=str, help='Input file, containing two columns srx and srr')
    parser.add_argument('-o', '--outdir', required=True, type=str, help="Path to result output directory of main process.")
    args = parser.parse_args()
    done_samples = check_done_sample(args.outdir)
    feature_store(done_samples, args.outdir)





import os
import re
from collections import namedtuple
from xopen import xopen
from io import BytesIO
from itertools import zip_longest
from typing import Optional, TextIO, Union
from more_itertools import grouper


class UnequalNumberReadsException(Exception):
    """Basic exception when PE data don't have the same number of reads"""


class MixedUpReadsException(Exception):
    """Basic exception when PE reads are not in the same order"""


Read = namedtuple("Read", "h1,seq,h2,qual")


def _strip_control_characters(string: str):
    """Removes ASCII control characters

    I have come across examples when there are control characters embedded
    in a read. This removes those before comparing lengths.
    """
    return "".join([x for x in string if ord(x) > 32])


class Fastq:
    """A simple FASTQ Parser"""
    def __init__(self, R1: str, R2: Optional[str] = None):
        self.R1 = R1
        self.R2 = R2
        self.libsize = None
        self.avgReadLen = None
        self.unequal_len = None
        self.bad_ecoding = None
        self.incomplete_read = None
        self.flags = set()

    def process(self):
        """Process FASTQ File
        Iterate over the FASTQ file returning each read as a string. If the
        FASTQ is from an ABI Solid sequencer it will raise an `AbiException`.
        The script also removes reads that have:
            - Are not encoded as ASCII
            - Have unequal length Sequence and Quality
            - Incomplete reads
        The object accumulates the Library Size (libsize) and the Average
        Read Length (avgReadLen). It also provides counts for the above
        problems.
        Yields
        ======
        String representation of each read.
        Raises
        ======
        NoReadsException: If FASTQ is empty.
        AbiException: If FASTQ is from ABI Solid.
        """
        if self._is_abi():  # Ignore samples stored as colorspace
            self.flags.add("abi_solid")
            return

        if self._is_empty(self.R1) & self._is_empty(self.R2):  # No data
            self.flags.add("download_bad")
            return

        if "keep_R1" in self.flags:  # Re-running b/c PE failed
            self.R2 = None
            yield from self._process_single_end()
            return

        if "keep_R2" in self.flags:  # Re-running b/c PE failed
            self.R1 = self.R2
            self.R2 = None
            yield from self._process_single_end()
            return

        if ~self._is_empty(self.R1) & (self.R2 is None):  # Single End (SE)
            self.flags.add("SE")
            yield from self._process_single_end()
            return

        if ~self._is_empty(self.R1) & ~self._is_empty(self.R2):  # Paired End (PE)
            self.flags.add("PE")
            yield from self._process_pair_end()
            return

        if ~self._is_empty(self.R1) & self._is_empty(self.R2):  # R2 is empty treat as SE
            self.flags.add("keep_R1")
            self.R2 = None
            yield from self._process_single_end()
            return

        if self._is_empty(self.R1) & ~self._is_empty(self.R2):  # R1 is empty treat as SE
            self.flags.add("keep_R2")
            self.R1 = self.R2
            self.R2 = None
            yield from self._process_single_end()
            return

    def open_fastq(self, fastq: Union[str, bytes] = None):
        if fastq is None:
            fastq = self.R1

        if isinstance(fastq, bytes):
            return BytesIO(fastq)
        elif fastq.endswith(".gz"):
            return xopen(fastq, "rb")
        elif fastq.endswith(".fastq") or fastq.endswith(".fq"):
            return open(fastq, "rb")
        elif isinstance(fastq, str):
            return BytesIO(fastq.encode("ascii"))

    def iter_reads(self, file_handle: TextIO) -> Read:
        for h1, seq, h2, qual in grouper(file_handle, 4):
            if seq is None and qual is None:
                continue
            yield Read(
                self._decode(h1), self._decode(seq), self._decode(h2), self._decode(qual),
            )

    def _process_single_end(self):
        self.libsize = 0
        self.unequal_len = 0
        self.bad_ecoding = 0
        self.incomplete_read = 0
        total_len = 0

        with self.open_fastq(self.R1) as fh:
            for read in self.iter_reads(fh):
                if self._is_incomplete(read):
                    # This should only happen if file was truncated.
                    self.incomplete_read += 1
                    continue

                if self._is_wrong_encoding(read):
                    self.bad_ecoding += 1
                    continue

                if self._is_unequal_seq_qual(read):
                    self.unequal_len += 1
                    continue

                self.libsize += 1
                total_len += len(read.seq)
                yield self._read_to_bytes(read)

        self.avgReadLen = total_len / self.libsize

    def _process_pair_end(self):
        self.libsize = 0
        self.unequal_len = 0
        self.bad_ecoding = 0
        self.incomplete_read = 0
        total_len = [0, 0]

        with self.open_fastq(self.R1) as fh1, self.open_fastq(self.R2) as fh2:
            for read1, read2 in zip_longest(self.iter_reads(fh1), self.iter_reads(fh2)):
                if read1 is None:
                    self.flags.remove("PE")
                    self.flags.add("keep_R2")
                    raise UnequalNumberReadsException

                if read2 is None:
                    self.flags.remove("PE")
                    self.flags.add("keep_R1")
                    raise UnequalNumberReadsException

                if self._is_incomplete(read1) | self._is_incomplete(read2):
                    # This should only happen if file was truncated.
                    print("incomplete")
                    self.incomplete_read += 1
                    continue

                if self._is_wrong_encoding(read1) | self._is_wrong_encoding(read2):
                    print("wrong")
                    self.bad_ecoding += 1
                    continue

                if self._is_unequal_seq_qual(read1) | self._is_unequal_seq_qual(read2):
                    print("unequal")
                    self.unequal_len += 1
                    continue

                if self._is_different_header(read1, read2):
                    print("different")
                    self.flags.remove("PE")
                    if os.stat(self.R1).st_size < os.stat(self.R2).st_size:
                        self.flags.add("keep_R2")
                    else:
                        self.flags.add("keep_R1")
                    raise MixedUpReadsException

                self.libsize += 1
                total_len[0] += len(read1.seq)
                total_len[1] += len(read2.seq)
                yield self._read_to_bytes(read1), self._read_to_bytes(read2)

        self.avgReadLen = [total_len[0] / self.libsize, total_len[1] / self.libsize]

    @staticmethod
    def _decode(value: Optional[bytes]):
        try:
            return value.decode("ascii", errors="ignore").strip()
        except AttributeError:
            return None

    @staticmethod
    def _is_empty(data: Union[str, bytes]) -> bool:
        if data is None:
            return True

        if isinstance(data, bytes) & len(data) > 0:
            return False

        if isinstance(data, str):
            if data.endswith(".gz") | data.endswith(".fq") | data.endswith(".fastq"):
                stat = os.stat(data)
                if stat.st_size > 1000:
                    return False
            elif len(data) > 0:
                return False

        return True

    def _is_abi(self):
        with self.open_fastq(self.R1) as fh:
            for read in self.iter_reads(fh):
                if self._is_abi_read(read):
                    return True
                break

        if self.R2:
            with self.open_fastq(self.R2) as fh:
                for read in self.iter_reads(fh):
                    if self._is_abi_read(read):
                        return True
                    break
        return False

    @staticmethod
    def _is_abi_read(read: Read) -> bool:
        """Look at read and determine if using abi colorspace.
        An Abi Solid read has the format:
            @SRR######.1 solid0527_####### length=50
            T120202210232000020002.00301000012.100...00........
            +SRR######.1 solid0527_####### length=50
            !/<%2/:%*)-%%0'--'')/.!%('1'%),+/%!&',!!!'+!!!!!!!!
        """
        if re.match(r"^T[\d\.]+$", read.seq):
            return True
        return False

    @staticmethod
    def _is_incomplete(read: Read) -> bool:
        """Determine if read has qual score"""
        if read.qual is None:
            return True
        return False

    def _is_wrong_encoding(self, read: Read) -> bool:
        """Determine if read is correctly encoded"""
        if self._is_invaid_header(read.h1) | self._is_invaid_header(read.h1):
            return True

        if self._is_invalid_seq(read.seq):
            return True

        if self._is_invalid_qual(read.qual):
            return True

        return False

    @staticmethod
    def _is_invaid_header(string: str):
        """Make sure only contains valid ascii characters
            Valid characters are between ascii 32-126. Where chr(32) is space
            and chr(126) is ~. Below are all the valid characters.
             !"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~
        """
        for character in string:
            if ord(character) < 32 | 126 < ord(character):
                return True
        return False

    @staticmethod
    def _is_invalid_seq(string: str):
        """Make sure sequence is a nucleotide or N."""
        for character in string:
            if character.upper() not in "ACTGN":
                return True
        return False

    @staticmethod
    def _is_invalid_qual(string: str):
        """Make sure only contains valid ascii characters
            Valid characters are between ascii 33-126. A quality score cannot
            contain a space (i.e., 32). Below are all the valid characters.
            !"#$%&'()*+,-./0123456789:;<=>?@ABCDEFGHIJKLMNOPQRSTUVWXYZ[\]^_`abcdefghijklmnopqrstuvwxyz{|}~
            See https://en.wikipedia.org/wiki/FASTQ_format for more details.
        """
        for character in string:
            if ord(character) < 33 or 126 < ord(character):
                return True
        return False

    @staticmethod
    def _is_unequal_seq_qual(read: Read) -> bool:
        """Determine if read has equal length seq and qual"""
        if len(read.seq) != len(read.qual):
            return True
        return False

    @staticmethod
    def _is_different_header(r1: Read, r2: Read) -> bool:
        """Determine if PE has same header"""
        # Ignore read length because they can be different
        regex = re.compile(r"length=\d+")
        r1_h1 = re.sub(regex, "", r1.h1)
        r2_h1 = re.sub(regex, "", r2.h1)
        if len(r1_h1)!=len(r2_h1):
            return True
        dif_words = []
        for m, n in zip(r1_h1, r2_h1):
            if m!=n:
                dif_words.append(m)
                dif_words.append(n)
        if len(dif_words)>2:
            return True
        return False

    @staticmethod
    def _read_to_bytes(read: Read) -> str:
        return f"{read.h1}\n{read.seq}\n{read.h2}\n{read.qual}\n".encode("ascii")

    def __str__(self):
        if isinstance(self.avgReadLen, list):
            avgReadLen = f"{self.avgReadLen[0] or 0.0:.0f}|{self.avgReadLen[1] or 0.0:.0f}"
        else:
            avgReadLen = f"{self.avgReadLen or 0.0:.0f}"

        return (
            f"{'R1':<20} : {self.R1}\n"
            f"{'R2':<20} : {self.R2}\n"
            f"{'# Reads':<20} : {self.libsize or 0:,}\n"
            f"{'# Avg Read Length':<20} : {avgReadLen}\n"
            f"{'# Seq/Qual Mismatch':<20} : {self.unequal_len or 0:,}\n"
            f"{'# Non-ASCII':<20} : {self.bad_ecoding or 0:,}\n"
            f"{'# Incomplete':<20} : {self.incomplete_read or 0:,}\n"
            f"{'Flags':<20} : {', '.join(self.flags)}\n"
        )


import sys, argparse

import numpy as np
import pandas as pd
from pathlib import Path

from .iforest import SraIsolationForest
#from .plot import outlier_umap, plot_importance

RANDOM_STATE = np.random.RandomState(42)


def detection(features_file):
    """Isolation forest training using features extracted from RNA seq"""
    rnaseq_features = pd.read_parquet(features_file)
    iso = SraIsolationForest(
        rnaseq_features, random_state=RANDOM_STATE, iso_kwargs=dict(n_estimators=100)
    )

    # Check that proportion of outliers in Train and Test is similar.
    assert np.isclose(iso.prop_outliers_test, iso.prop_outliers_train, atol=0.01)

    # Save a list of good rnaseq samples
    inliers = iso.inliers(rnaseq_features).index.tolist()
    outliers = iso.outliers(rnaseq_features).index.tolist()
    # plot_importance(iso, features_file)
    # outlier_umap(features_file, inliers)
    rnaseq_features['labels'] = rnaseq_features.index.to_series().apply(
        lambda x: "Inliers" if x in inliers else "Outlier"
    )
    result_path = Path(features_file).parent.parent / "result.csv"
    rnaseq_features.to_csv(result_path)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument('-f', '--features_file', required=True, type=str,
                        help="Path to feature dataframe file.")
    args = parser.parse_args()
    detection(args.features_file)




import argparse
import configparser
import logging
import os

import pandas as pd

logging.basicConfig(format='%(asctime)s %(message)s')
logger = logging.getLogger("MassiveQC")
logger.setLevel(logging.WARNING)
from pathlib import Path
from concurrent import futures
from tqdm import tqdm

# sys.path.insert(0, "/home/mwshi/github/MassiveQC")
from .get_sra import get_sra
from .check_fq import check_fq
from .fastq_screen import fastq_screen
from .atropos import atropos
from .hisat2 import Hisat2
from .collectrnaseqmetrics import CollectRnaseqMetrics
from .markduplicates import MarkDuplicates
from .FeatureCounts import FeatureCounts
from .feature_store import check_done_sample, feature_store
from .detection import detection
from .parser import remove_file

def init_wd():
    Path(outdir).mkdir(exist_ok=True)
    Path(download_path).mkdir(exist_ok=True)
    feature_path.mkdir(exist_ok=True)
    (feature_path / "layout").mkdir(exist_ok=True)
    Path(QC_dir).mkdir(exist_ok=True)
    (feature_path / "fastq_screen").mkdir(exist_ok=True)
    (feature_path / "atropos").mkdir(exist_ok=True)
    Path(Bam_dir).mkdir(exist_ok=True)
    (feature_path / "hisat2").mkdir(exist_ok=True)
    (feature_path / "aln_stats").mkdir(exist_ok=True)
    (feature_path / "strand").mkdir(exist_ok=True)
    (feature_path / "rnaseqmetrics").mkdir(exist_ok=True)
    (feature_path / "genebody_coverage").mkdir(exist_ok=True)
    (feature_path / "markduplicates").mkdir(exist_ok=True)
    (feature_path / "count_summary").mkdir(exist_ok=True)
    Path(Count_dir).mkdir(exist_ok=True)
    (feature_path / "DoneSample").mkdir(exist_ok=True)


def process(SRR):
    logger.info(f"Start download {SRR}")
    if not skip_download:
        fq_mode = get_sra(SRR, download_path, ascp_key)
        logger.info(f"Complete download {SRR}")
    if only_download:
        return

    # check_fq
    # Check if the result file exists.
    logger.info(f"Start check {SRR} fastq file")
    summary_file = feature_path / "layout" / f"{SRR}.parquet"
    if summary_file.exists():
        logger.info(f"{SRR} fastq file has been checked")
    else:
        raw_fqs = check_fq(SRR, download_path, QC_dir, feature_path)
        # remove the raw fastq
        if remove_fastq:
            for k in raw_fqs:
                logger.info(f"Remove {k}")
                remove_file(k)

    # fastq_screen
    fastq_screen_output = feature_path / "fastq_screen" / f"{SRR}.parquet"
    if fastq_screen_output.exists():
        logger.info(f"{SRR} fastq_screen step has been done")
    else:
        fastq_screen(SRR, QC_dir, feature_path.as_posix(), fastq_screen_config, THREADS)

    # atropos
    atropos_output = feature_path / "atropos" / f"{SRR}.parquet"
    if atropos_output.exists():
        logger.info(f"{SRR} atropos step has been done")
    else:
        atropos(feature_path.as_posix(), SRR, QC_dir, THREADS)

    # hisat2
    _hisat2 = Path(feature_path) / "hisat2" / f"{SRR}.parquet"
    _alnStat = Path(feature_path) / "aln_stats" / f"{SRR}.parquet"
    if _hisat2.exists() and _alnStat.exists():
        logger.info(f"{SRR} hisat2 step has been done")
    else:
        hisat_runner = Hisat2(feature_path.as_posix(), SRR, QC_dir, Bam_dir, THREADS, reference, splice=splice)
        trim_fqs = hisat_runner.hisat2()
        if remove_fastq:
            for k in trim_fqs:
                logger.info(f"Remove {k}")
                remove_file(k)

    # metrics
    strand = feature_path / "strand" / f"{SRR}.parquet"
    table = feature_path / "rnaseqmetrics" / f"{SRR}.parquet"
    coverage = feature_path / "genebody_coverage" / f"{SRR}.parquet"
    if strand.exists() and table.exists() and coverage.exists():
        logger.info(f"{SRR} collectrnaseqmetrics step has been done")
    else:
        metrics_runner = CollectRnaseqMetrics(feature_path.as_posix(), SRR, Bam_dir, THREADS, ref_flat, picard)
        metrics_runner.collectrnaseqmetrics()

    # markduplicates
    markdup = feature_path / "markduplicates" / f"{SRR}.parquet"
    if markdup.exists():
        logger.info(f"{SRR} markduplicates step has been done")
    else:
        markdup_runner = MarkDuplicates(feature_path.as_posix(), SRR, Bam_dir, THREADS, picard)
        markdup_runner.markduplicates()

    # FeatureCounts
    count_summary = feature_path / "count_summary" / f"{SRR}.parquet"
    if count_summary.exists():
        logger.info(f"{SRR} FeatureCounts step has been done")
    else:
        count_runner = FeatureCounts(feature_path.as_posix(), SRR, Bam_dir, Count_dir, gtf, THREADS)
        bam_file = count_runner.FeatureCounts()
        if remove_bam:
            remove_file(bam_file)

    # complete one srr, touch one file
    (feature_path / "DoneSample" / SRR).touch()
    return SRR


def local_thread(SRRs):
    init_wd()
    down_samples = os.listdir(feature_path / "DoneSample")
    pre_SRRs = [x for x in SRRs if x not in down_samples]
    with futures.ThreadPoolExecutor(max_workers=workers) as executor:
        tasks = [executor.submit(process, srr) for srr in pre_SRRs]
        for r in tqdm(futures.as_completed(tasks), total=len(tasks)):
            if r.exception():
                logger.error("One sample failed")
            else:
                res = r.result()


def get_arguments():
    # parse the config file
    pre = argparse.ArgumentParser(add_help=False)
    pre.add_argument('-c', '--conf', type=str, required=False)
    pre_args, _ = pre.parse_known_args()
    config_args = {}
    if pre_args.conf:
        config = configparser.ConfigParser()
        with open(pre_args.conf, "r") as stream:
            config.read_string("[Defaults]\n" + stream.read())
        config_args.update(dict(config.items("Defaults")))
    parser = argparse.ArgumentParser(description='...', parents=[pre])
    parser.add_argument('-i', '--input', required=True, type=str, help='Input file, containing two columns srx and srr')
    parser.add_argument('-a', '--ascp_key', type=str, help='Locate aspera key. Default $HOME/.aspera/connect/etc/asperaweb_id_dsa.openssh')
    parser.add_argument('-f', '--fastq_screen_config', required=True, type=str, help="Path to the fastq_screen conf file, can be download from fastq_screen website")
    parser.add_argument('-g', '--gtf', required=True, type=str, help="Path to the GTF file with annotations")
    parser.add_argument('-x', '--ht2-idx', dest="ht2_idx", required=True, type=str, help="Hisat2 index filename prefix")
    parser.add_argument('-k', '--known-splicesite-infile', dest="known_splicesite_infile", type=str, help="Hisat2 splicesite file, provide a list of known splice sites")
    parser.add_argument('-p', '--picard', required=True, type=str, help="Path to picard.jar")
    parser.add_argument('-r', '--ref_flat', required=True, type=str, help="Path to refflat file")
    parser.add_argument('-o', '--outdir', required=True, type=str, help="Path to result output directory. If it doesn't exist, it will be created automatically")
    parser.add_argument('-w', '--workers', type=int, help="The number of simultaneous tasks", default=2)
    parser.add_argument('-t', '--THREADS', type=int, help="The number of threads for tools like Hisat2 in one task", default=4)
    parser.add_argument('-d', '--download', type=str, help="Path to SRA fastq files. The default is $OUTDIR/download")
    parser.add_argument('--only_download', action="store_true", help="Only run the download step", default=False)
    parser.add_argument('--skip_download', action="store_true", help="Skip the download step", default=False)
    parser.add_argument('--remove_fastq', action="store_true", help="Don't remain the fastq after running hisat2", default=False)
    parser.add_argument('--remove_bam', action="store_true", help="Don't remain the bam after running FeatureCounts", default=False)
    parser.set_defaults(**config_args)
    if pre_args.conf:
        for action in parser._actions:
            if action.dest in config_args:
                action.required = False

    return parser.parse_args()


def main():
    args = get_arguments()
    global only_download
    only_download = args.only_download
    global skip_download
    skip_download = args.skip_download
    global remove_fastq
    remove_fastq = args.remove_fastq
    global remove_bam
    remove_bam = args.remove_bam
    input_file = args.input.strip('"')
    global ascp_key
    ascp_key = args.ascp_key.strip('"')
    global gtf
    gtf = args.gtf.strip('"')
    global fastq_screen_config
    fastq_screen_config = args.fastq_screen_config.strip('"')
    global reference
    reference = args.ht2_idx.strip('"')
    global splice
    splice = args.known_splicesite_infile.strip('"')
    global picard
    picard = args.picard.strip('"')
    global ref_flat
    ref_flat = args.ref_flat.strip('"')
    global outdir
    outdir = args.outdir.strip('"')
    global workers
    workers = args.workers
    global THREADS
    THREADS = args.THREADS
    global download_path
    if args.download:
        download_path = args.download
    else:
        download_path = os.path.join(outdir, "download")
    global QC_dir
    QC_dir = os.path.join(outdir, "QC_dir")
    global Bam_dir
    Bam_dir = os.path.join(outdir, "Bam")
    global Count_dir
    Count_dir = os.path.join(outdir, "Count")
    global feature_path
    feature_path = os.path.join(outdir, "Features")
    feature_path = Path(feature_path)
    global done_sample
    done_sample = feature_path / "done_sample.txt"
    # init workshop
    init_wd()
    srr_df = pd.read_table(input_file, comment='#')
    if len(srr_df.columns) == 1:
        # only have srr column
        srr_df.columns = ["srr"]
    elif len(srr_df.columns) == 2:
        # have srx and srr column
        srr_df.columns = ["srx", "srr"]

    SRRs = srr_df["srr"].values.tolist()
    logger.info(f"Start processing, {len(SRRs)} srrs will be processed")
    # run process local
    local_thread(SRRs)
    if not only_download:
        done_samples = check_done_sample(outdir)
        feature_store(done_samples, outdir)
        detection((feature_path / "features.parquet").as_posix())


if __name__ == "__main__":
    main()



import logging
import os.path
from pathlib import Path
import pandas as pd
from .command import run_command
from typing import Optional, Tuple
from .parser import parse_picardCollect_summary, parse_picardCollect_hist, remove_file

logger = logging.getLogger("MassiveQC")


class CollectRnaseqMetrics(object):
    """This class uses picard to extract RnaSeqMetrics information"""
    def __init__(self, feature_path: str, SRR: str, Bam_dir: str,
                 THREADS: int, ref_flat: str, picard: str, MEM: Optional[int] = 3):
        self.feature_path = Path(feature_path)
        self.layout = Path(feature_path) / f"{SRR}.parquet"
        self.SRR = SRR
        self.Bam_dir = Path(Bam_dir)
        self.THREADS = THREADS
        self.MEM = MEM
        self.ref_flat = ref_flat
        self.picard = picard

    def collectrnaseqmetrics(self):
        bam = self.Bam_dir / f"{self.SRR}.sorted.bam"
        unstranded, first, second = self.run_picard(bam)
        self.summarize(unstranded, first, second)
        remove_file(unstranded.as_posix())
        remove_file(first.as_posix())
        remove_file(second.as_posix())
        logger.info(f"{self.SRR} Complete metrics")

    def run_picard(self, bam: Path):
        cmd = (
            f"java -Xmx{self.MEM}g "
            f"-jar {self.picard} CollectRnaSeqMetrics "
            f"REF_FLAT={self.ref_flat} "
            f"INPUT={bam} OUTPUT={{out_file}} "
            f"STRAND={{strand}}"
        )
        logger.info(f"{self.SRR} Start CollectRnaSeqMetrics")
        # Unstranded
        un_out = self.feature_path / f"{self.SRR}.unstranded.txt"
        result_un = run_command(cmd.format(out_file=un_out, strand="NONE"))
        self._check_log(result_un, self.SRR)

        # First stranded
        first_out = self.feature_path / f"{self.SRR}.first_stranded.txt"
        result_first = run_command(cmd.format(out_file=first_out, strand="FIRST_READ_TRANSCRIPTION_STRAND"))
        self._check_log(result_first, self.SRR)

        # Second stranded
        second_out = self.feature_path / f"{self.SRR}.second_stranded.txt"
        result_second = run_command(cmd.format(out_file=second_out, strand="SECOND_READ_TRANSCRIPTION_STRAND"))
        self._check_log(result_second, self.SRR)
        return un_out, first_out, second_out

    def summarize(self, unstranded: Path, first: Path, second: Path):
        # Parse the flags
        if self._parse_stranded(first):
            strand_ = "same_strand"
        elif self._parse_stranded(second):
            strand_ = "opposite_strand"
        else:
            strand_ = "unstranded"

        idx = pd.Index([self.SRR], name="srr")
        # Write strand flag
        strand = self.feature_path / "strand" / f"{self.SRR}.parquet"
        strand_df = pd.DataFrame([[strand_]], index=idx, columns=["strand"])
        strand_df.to_parquet(strand)

        # Parse main table
        table = self.feature_path / "rnaseqmetrics" / f"{self.SRR}.parquet"
        table_df = self._parse_table(unstranded)
        table_df.index = idx
        table_df.to_parquet(table)

        # Parse genome coverage histogram
        coverage = self.feature_path / "genebody_coverage" / f"{self.SRR}.parquet"
        coverage_df = parse_picardCollect_hist(unstranded)
        coverage_df.index = idx
        coverage_df.to_parquet(coverage)

    @staticmethod
    def _parse_stranded(file_name: Path) -> bool:
        return (parse_picardCollect_summary(file_name).PCT_CORRECT_STRAND_READS >= 0.75)[0]

    @staticmethod
    def _parse_table(file_name: Path) -> pd.DataFrame:
        df = parse_picardCollect_summary(file_name)[
            [
                "PCT_CODING_BASES",
                "PCT_UTR_BASES",
                "PCT_INTRONIC_BASES",
                "PCT_INTERGENIC_BASES",
                "PCT_MRNA_BASES",
                "MEDIAN_CV_COVERAGE",
                "MEDIAN_5PRIME_BIAS",
                "MEDIAN_3PRIME_BIAS",
                "MEDIAN_5PRIME_TO_3PRIME_BIAS",
            ]
        ].fillna(0.0)
        df.PCT_CODING_BASES = df.PCT_CODING_BASES * 100
        df.PCT_UTR_BASES = df.PCT_UTR_BASES * 100
        df.PCT_INTRONIC_BASES = df.PCT_INTRONIC_BASES * 100
        df.PCT_INTERGENIC_BASES = df.PCT_INTERGENIC_BASES * 100
        df.PCT_MRNA_BASES = df.PCT_MRNA_BASES * 100

        df.columns = [x.lower() for x in df.columns]
        df.columns = [x.replace("pct_", "percent_") for x in df.columns]
        return df

    @staticmethod
    def _check_log(log: str, SRR) -> None:
        if "CollectRnaSeqMetrics done" not in log:
            raise PicardException(f"{SRR} not complete")


class PicardException(Exception):
    """Picard Processing Exception"""



from setuptools import setup, find_packages

# with open('requirements.txt') as f:
#     requirements = f.readlines()
requirements = [
    'scikit-learn',
    'shap',
    'pysradb',
    'xopen',
    'NumPy',
    'Pandas >=1.3.2',
    'fastparquet',
    'more-itertools',
    'tqdm'
]

with open('README.md') as f:
    long_description = f.read()


setup(
    name='MassiveQC',
    version='0.0.6',
    author='shimw6828',
    author_email='shimw6828@qq.com',
    url='https://github.com/shimw6828/MassiveQC',
    description='Tools for QC massive RNA-seq samples',
    long_description=long_description,
    long_description_content_type='text/markdown',
    license='MIT',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'MultiQC = MassiveQC.MultiProcess:main',
            'SingleQC = MassiveQC.SingleProcess:main',
            'IsoDetect = MassiveQC.IsoDetect:main'
        ]
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: POSIX",
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
    install_requires=requirements,
    python_requires='>=3.7'
)


import subprocess


def run_command(command, verbose=False):
    """Run a shell command"""
    process = subprocess.Popen(
        command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT, shell=True
    )
    logout = ""
    while True:
        output = process.stdout.readline().strip()
        output = output.decode("utf-8")
        if output == "" and process.poll() is not None:
            break
        if output:
            logout += str(output.strip()) + "\n"
            if verbose:
                print((str(output.strip())))
    rc = process.poll()
    return logout


import os
from xopen import xopen
import logging
from pathlib import Path
from typing import Optional
from .fastq import Fastq, MixedUpReadsException, UnequalNumberReadsException
import pandas as pd
from .parser import remove_file

logger = logging.getLogger("MassiveQC")


class DownloadException(Exception):
    """Basic exception for problems downloading from SRA"""


class AbiException(Exception):
    """Basic exception when ABI file was downloaded from SRA"""


def check_and_compress_fastq(r1: str, QC_dir: str, r2: Optional[str] = None):
    """Check the reads quality and compress them in a new fastq file

    **parameter**

    r1: str
        The first fastq file

    QC_dir: str
        The directory of result fastq file after check_fq.

    r2: str or None
        The second fastq file

    **return**

    r1_gz: str

    r2_gz: str

    fq: Fastq class
    """
    fq = Fastq(r1, r2)
    r1_gz = os.path.join(QC_dir, os.path.basename(r1))
    if r2 is None:
        logger.info("Processing FASTQ as Single-End")
        run_as_se(fq, r1_gz)
        return r1_gz, fq
    else:
        logger.info("Processing FASTQ as Pair-End")
        r2_gz = os.path.join(QC_dir, os.path.basename(r2))
        run_as_pe(fq, r1_gz, r2_gz)
        return r1_gz, r2_gz, fq


def run_as_se(fq: Fastq, R1_out: str) -> None:
    """Check the reads quality and compress for single-end RNA-seq

    **parameter**

    fq: Fastq
        Fastq class

    R1_out: Path
        The result R1 fastq path

    **return**

    None

    """
    with xopen(R1_out, "wb") as file_out1:
        for read in fq.process():
            file_out1.write(read)
    if "abi_solid" in fq.flags:
        remove_file(R1_out)
        raise AbiException
    if "download_bad" in fq.flags:
        remove_file(R1_out)
        raise DownloadException("Empty FASTQ")
    if fq.libsize < 100_000:
        remove_file(R1_out)
        raise DownloadException("<100,000 reads")



def run_as_pe(fq: Fastq, R1_out: str, R2_out: str) -> None:
    """Check the reads quality and compress for single-end RNA-seq
    **parameter**

    fq: Fastq
        Fastq class

    R1_out: Path
        The result R1 fastq path

    R2_out: Path
        The result R2 fastq path

    **return**

    None
    """
    try:
        with xopen(R1_out, "wb") as file_out1, xopen(R2_out, "wb") as file_out2:
            for read1, read2 in fq.process():
                file_out1.write(read1)
                file_out2.write(read2)
        if "abi_solid" in fq.flags:
            raise AbiException
        if "download_bad" in fq.flags:
            raise DownloadException("Empty FASTQ")
        if fq.libsize < 100_000:
            raise DownloadException("<100,000 reads")
    except (UnequalNumberReadsException, MixedUpReadsException):
        remove_file(R1_out)
        remove_file(R2_out)
        if "keep_R1" in fq.flags:
            run_as_se(fq, R1_out)
        elif "keep_R2" in fq.flags:
            run_as_se(fq, R2_out)


def save_output(feature_path, fq, SRR):
    """Save the output. This method will create a layout file contain 4 columns
    **parameter**

    feature_path: str
        The Feature dir path.

    fq: Fastq
        Fastq class.

    SRR: str
        SRR ID.

    **return**

    None
    """
    summary_file = os.path.join(feature_path, "layout", f"{SRR}.parquet")
    layout = fq.flags.intersection(set(["SE", "PE", "keep_R1", "keep_R2"])).pop()
    idx = pd.Index([SRR], name="srr")
    if isinstance(fq.avgReadLen, list):
        r1, r2 = fq.avgReadLen
    else:
        r1, r2 = fq.avgReadLen, 0.0
    df = pd.DataFrame([[layout, fq.libsize, r1, r2]], index=[idx],
                      columns=["layout", "libsize", "avgLen_R1", "avgLen_R2"])
    df.to_parquet(summary_file)


def run_check_fq(SRR, SRA_path, QC_dir, feature_path):
    """Determine the layout of RNA-seq and Run check_and_compress_fastq
    **parameter**
    SRR: str
        SRR ID.
    SRA_path: str
        The path of fastq download directory.
    QC_dir: str
       The directory of result fastq file after check_fq.
    feature_path: str
        The directory of Feartures

    **return**
    None
    """
    file_list = [file for file in os.listdir(SRA_path) if file.startswith(SRR)]
    raw_fqs = []
    summary_file = os.path.join(feature_path, "layout", f"{SRR}.parquet")
    if len(file_list) == 2:
        logger.info("Pair-End QC")
        r1 = Path(SRA_path, f"{SRR}_1.fastq.gz")
        r2 = Path(SRA_path, f"{SRR}_2.fastq.gz")
        if os.path.exists(summary_file):
            return [r1.as_posix(), r2.as_posix()]
        r1_gz, r2_gz, fq = check_and_compress_fastq(r1=r1.as_posix(), QC_dir=QC_dir, r2=r2.as_posix())
        raw_fqs.append(r1.as_posix())
        raw_fqs.append(r2.as_posix())
    elif len(file_list) == 1:
        logger.info("Single-End QC")
        r1 = Path(SRA_path, f"{SRR}.fastq.gz")
        if os.path.exists(summary_file):
            return [r1.as_posix()]
        r1_gz, fq = check_and_compress_fastq(r1=r1.as_posix(), QC_dir=QC_dir)
        raw_fqs.append(r1.as_posix())
    elif len(file_list) == 0:
        raise DownloadException
    save_output(feature_path, fq, SRR)
    return raw_fqs


def check_fq(SRR, SRA_path, QC_dir, feature_path):
    """This is the main function for checking the reads quality

    **parameter**

    SRR: str
        SRR ID.

    SRA_path: str
        The path of fastq download directory.

    QC_dir: str
       The directory of result fastq file after check_fq.

    feature_path: str
        The directory of Features.
    """
    try:
        raw_fqs = run_check_fq(SRR, SRA_path, QC_dir, feature_path)
        logger.info(f"Complete check {SRR} fastq file")
        return raw_fqs
    except AbiException:
        logger.warning(f"Flagging {SRR} as ABI Solid")
        raise AbiException
    except DownloadException as error:
        logger.warning(f"Flagging {SRR} as Download Bad: {error}")
        raise DownloadException



import logging
from pathlib import Path
import pandas as pd
from .command import run_command
from typing import Optional, Tuple
from .parser import parse_hisat2, parse_samtools_stats, parse_bamtools_stats, remove_file

logger = logging.getLogger("MassiveQC")


class Hisat2(object):
    """This class can run hisat2 and extract the alignment summary"""
    def __init__(self, feature_path: str, SRR: str, QC_dir: str, Bam_dir: str,
                 THREADS: int, reference: str, strand: Optional[str] = None,
                 splice: Optional[str] = None):
        self.feature_path = Path(feature_path)
        self.layout = Path(feature_path) / "layout" / f"{SRR}.parquet"
        self.SRR = SRR
        self.QC_dir = Path(QC_dir)
        self.Bam_dir = Path(Bam_dir)
        self.THREADS = THREADS
        self.reference = reference
        self.strand = strand
        self.splice = splice
        self.r1 = None
        self.r2 = None

    def hisat2(self):
        logger.info(f"")
        layout_ = pd.read_parquet(self.layout).layout[0]
        trim_fqs = []
        if layout_ == "PE":
            self.r1 = self.QC_dir / f"{self.SRR}_1.trim.fastq.gz"
            self.r2 = self.QC_dir / f"{self.SRR}_2.trim.fastq.gz"
        elif layout_ == "Keep_R1":
            self.r1 = self.QC_dir / f"{self.SRR}_1.trim.fastq.gz"
            r2 = None
        elif layout_ == "Keep_R2":
            self.r1 = self.QC_dir / f"{self.SRR}_2.trim.fastq.gz"
            self.r2 = None
        else:
            self.r1 = self.QC_dir / f"{self.SRR}.trim.fastq.gz"
            self.r2 = None
        results, sam = self.run_hisat2()
        # return results
        _hisat2 = self.feature_path / "hisat2" / f"{self.SRR}.parquet"
        self.check_hisat(results, _hisat2, self.SRR)
        bam, bai = self.compress_sort_and_index(sam)
        remove_file(sam.as_posix())
        _alnStat = self.feature_path / "aln_stats" / f"{self.SRR}.parquet"
        self.alignment_stats(bam, _alnStat)
        trim_fqs.append(self.r1.as_posix())
        trim_fqs.append(self.r2.as_posix())
        return trim_fqs

    def run_hisat2(self):
        if self.r2:
            fastqs = f"-1 {self.r1} -2 {self.r2}"
        else:
            fastqs = f"-U {self.r1}"
        # Look up strand if it is there
        strand_param = ""
        if self.strand:
            strand_ = pd.read_parquet(self.strand).strand[0]
            if (self.layout_ == "PE") & (strand_ == "first_strand"):
                strand_param = "--rna-strandness FR"
            elif (self.layout_ == "PE") & (strand_ == "second_strand"):
                strand_param = "--rna-strandness RF"
            elif strand_ == "first_strand":
                strand_param = "--rna-strandness F"
            elif strand_ == "second_strand":
                strand_param = "--rna-strandness R"
        splice_param = ""
        if self.splice:
            splice_param = f"--known-splicesite-infile {self.splice}"
        sam = self.Bam_dir / f"{self.SRR}.sam"
        cmd = (
            "hisat2 "
            f"-x {self.reference} "
            f"{fastqs} "
            f"--threads {self.THREADS} "
            "--max-intronlen 300000 "
            f"{strand_param} "
            f"{splice_param} "
            f"-S {sam} "
        )
        logger.info(f"{self.SRR} Start Hisat2 alignment")
        logger.info(f"{cmd}")
        results = run_command(cmd)
        logger.info(f"{self.SRR} Complete Hisat2 alignment")
        return results, sam

    def compress_sort_and_index(self, sam: Path) -> Tuple[Path, Path]:
        sorted_bam = self.Bam_dir / f"{self.SRR}.sorted.bam"
        sorted_bai = self.Bam_dir / f"{self.SRR}.sorted.bam.bai"
        cmd = f"samtools view -Sb -q 20 --threads {self.THREADS} {sam} " \
              f"|samtools sort -l 9 --output-fmt BAM " \
              f" --threads {self.THREADS} -o {sorted_bam} && " \
              f"samtools index {sorted_bam}"
        logger.info(cmd)
        results = run_command(cmd)
        if "error" in results.lower():
            logger.warning(f"{self.SRR} samtoots error")
            raise Exception(f"samtoots error")
        else:
            return sorted_bam, sorted_bai

    def alignment_stats(self, bam: Path, output_file: Path):
        logger.info(f"samtools stats {bam}")
        result1 = run_command(f"samtools stats {bam}", verbose=False)
        logger.info(f"bamtools stats -in {bam}")
        result2 = run_command(f"bamtools stats -in {bam}", verbose=False)
        # Summarize
        df = pd.concat([self._samtools(result1), self._bamtools(result2)], axis=1, sort=False)
        df.index = pd.Index([self.SRR])
        df.to_parquet(output_file)

    @staticmethod
    def check_hisat(log: str, output_file: str, srr) -> None:
        """Example Hisat2 output
            # Example SE
            83440 reads; of these:
            83440 (100.00%) were unpaired; of these:
                3605 (4.32%) aligned 0 times
                76302 (91.45%) aligned exactly 1 time
                3533 (4.23%) aligned >1 times
            95.68% overall alignment rate
            # Example PE
            1338 reads; of these:
            1338 (100.00%) were paired; of these:
                468 (34.98%) aligned concordantly 0 times
                839 (62.71%) aligned concordantly exactly 1 time
                31 (2.32%) aligned concordantly >1 times
                ----
                468 pairs aligned concordantly 0 times; of these:
                22 (4.70%) aligned discordantly 1 time
                ----
                446 pairs aligned 0 times concordantly or discordantly; of these:
                892 mates make up the pairs; of these:
                    807 (90.47%) aligned 0 times
                    74 (8.30%) aligned exactly 1 time
                    11 (1.23%) aligned >1 times
            69.84% overall alignment rate
        """
        df = parse_hisat2(log).fillna(0)
        df.index = pd.Index([srr])
        df.to_parquet(output_file)
        uniquely_aligned = (
                df.iloc[0, :]["num_concordant_reads_uniquely_aligned"]
                + df.iloc[0, :]["num_uniquely_aligned"]
        )

        per_aligned = df.iloc[0, :]["per_alignment"]

        if (per_aligned < 1) | (uniquely_aligned < 1000):
            raise Exception(f"Poor alignment: {uniquely_aligned:,} ({per_aligned}%)")

    @staticmethod
    def _samtools(stats_file: Path) -> pd.DataFrame:
        return parse_samtools_stats(stats_file)[
            [
                "reads_MQ0",
                "average_quality",
                "insert_size_average",
                "insert_size_standard_deviation",
                "inward_oriented_pairs",
                "outward_oriented_pairs",
                "pairs_with_other_orientation",
                "pairs_on_different_chromosomes",
            ]
        ]

    @staticmethod
    def _bamtools(stats_file: Path) -> pd.DataFrame:
        return parse_bamtools_stats(stats_file)[["Percent Forward", "Percent Reverse"]]
